--[[
Name: FlexBar 2
Revision: $Rev: SVN $
Author(s): Cncfanatics (cvcncfanatic@runemasters.net)
    Dhaggar (isheridan@savagevines.com)
Website: http://flexdev.mmorpgaming.com
Documentation: http://flexdev.mmorpgaming.com/index.php?n=Main.Docs
Description: Makes your 120 actionbuttons incredibly versatile
Dependencies: AceLibrary
]]
FlexBar2.Controller = {};

-- Localize Model & Controller for increased performance
local Model = FlexBar2.Model;
local Controller = FlexBar2.Controller;

Controller.DefaultEventList = { "ACTIONBAR_SLOT_CHANGED", "UNIT_INVENTORY_CHANGED", "FB2_Startup", "FB2_PostLoad", "FB2_ActionIsUsableStateChanged",}; --Default event list, when these events are triggered they get passed to any function that matches the event name in Controller.Events table (*not* in Controller.Events.Button) (Note: non blizzard events used to be declared in another table, but I just merged both tables)
Controller.DefaultButtonEventList = { "ACTIONBAR_UPDATE_COOLDOWN", "UPDATE_KEYBINDINGS", "UNIT_INVENTORY_CHANGED", "UPDATE_SHAPESHIFT_FORM", "UPDATE_STEALTH"}; -- Default *button* event list, these behave specialy, instead of being registered once, they get registered for each shown button frame, and thus, the functions in the Controller.Events.Button table get called once per shown button

function Controller:Startup() -- Gets called with the PLAYER_LOGIN event from the FlexBar2.lua file, this is late enough to have access to all UI info, but soon enough to not experience any in-combat lockdown yet
	-- Localizing functions that are called way too often for performance (Note: because of this self will be appended to the arg list when calling the function)
	local AddCustomEvent = Controller.AddCustomEvent;
	local GenerateClass = Controller.GenerateClass;
	local RegisterEvent = FlexBar2.RegisterEvent;


	if(Model:IsEmpty()) then -- IsEmpty() check on the database, if its empty, declare some *needed* default settings
		FlexBar2:Debug("Empty DB detected, setting defaults");
		Model:SetCustomSetting("MaxButtons", 120); -- Set normal ammount of buttons to 120, user can make more or less if he wants, but there still are only 120 action IDs
		Model:SetCustomSetting("SavedCustomEvents", {}); -- SavedCustomEvents table that contains all code saved through sessions (this should *not* be used by other addons, this is made for users) (see the SaveCustomEvent function declaration for details on usage of that function)
	end


	local HeaderFrame = CreateFrame("Frame", nil, UIParent, "SecureStateDriverTemplate");
	HeaderFrame:SetAttribute("statemap-stance", "$input");
	Model:SetTempData("HeaderFrame", HeaderFrame); --Setup the header frame that inherits from the SecureStateHeaderTemplate

	Model:SetTempData("CustomEventTable", {}); --This creates the so called CustomEventTable, this get populated with data from SavedCustomEvents (see the SaveCustomEvent function declaration for details on usage of that function) and can be populated further at will
	Model:SetTempData("TextSubTable", {}); -- This creates the TextSubTable, this doesn't get populated by saved data, but instead has to be populated manualy at each startup, so mods can declare their own textsubs at startup (but an OptionalDepency to FlexBar2), see the RegisterTextSub function declaration for more details
	local SavedCustomEvents = Model:GetCustomSetting("SavedCustomEvents"); --Get a local copy from the SavedCustomEvents
	for n,v in pairs(SavedCustomEvents) do -- Populate the CustomEventTable with the *strings* of code saved in SavedCustomEvents (more details in the SaveCustomEvent function declaration)
		--[[
			Note: Expected format (this is suposed to be handled by the SaveCustomEvent function and is put here just for reference): SavedCustomEvents =
			{
				EVENT_NAME =
				{
					[1] = "string(); of(); code();",
					[2] = "another(); string(); of(); code();",
				},
				FB2_CustomEventName =
				{
					[1] = "string(); of(); code();",
					[2] = "another(); string(); of(); code();",
				}
			}
		]]
		if(type(v) == "table") then -- Sanity check (we don't want errors spit out, i'll just output some text if I meet any errors)
			for _, script in ipairs(v) do
				AddCustomEvent(self, n, loadstring(script));
			end
		else -- Debug code in case the format of the table ain't valid
			FlexBar2:Print("Warning: Invalid format in the SavedCustomEvents table ! please fix, skipping the nonvalid event table (event name: '" .. n .. "')");
		end
	end


	for ButtonNumber = 1, Model:GetCustomSetting("MaxButtons") do -- Generate the ammount of buttons saved in the MaxButtons custom Setting (defaults to 120)
		GenerateClass(self, ButtonNumber); -- See the GenerateClass function declaration on more info on what this does
	end


		-- Register the "built-in" Text subs at each loggin (see the RegisterTextSub function declaration for more details on how to use this function)
	Controller:RegisterTextSub("cd", "OnUpdate", -- Cooldown numbers textsub, unfortunatly triggered with OnUpdate, I did not find any suitable event, if you know one, please tell me. Syntax: $(cd)
		function(class)
			if(type(Model:GetButtonSetting(class.ButtonID, "Action").action) ~= "number") then return ""; end -- Make sure the current button has an ActionID linked to it (might support spell cooldowns later, but that ain't a priority) (note on this check: the button does not have to be an actionbutton to have an actionid, u can have macro buttons or whatever with actionid's as well, they will then react to cooldowns and such as if they were that button)
			local start, duration, enable = GetActionCooldown(Model:GetButtonSetting(class.ButtonID, "Action").action); -- Regular GetActionCooldown call, nothing special here
			if(duration > 1.5) then -- Because of this line the global cooldown gets ignored (hardcoded atm, might add a config setting for it later)
				local TimeLeft = (duration-(GetTime()-start)); -- calculation for the time left for the button to be on cooldown
				if(TimeLeft >= 60) then -- if the time left is bigger then 60 its probably it won't fit on the button, so we display minutes instead of seconds (I shouldn't get problems for hours right ? never saw a cooldown > 30 minutes)
					return(math.floor ((TimeLeft / 60) + 0.5) .. "m"); -- Round & return
				else -- time left less then 60 seconds, we show seconds
					if(TimeLeft == 0) then -- timeleft 0, so we return an empty string
						return "";
					else
						return(math.floor (TimeLeft) .. "s"); -- Round & return
					end
				end
			end
			return ""; -- Fallback to return an empty string
		end);

		Controller:RegisterTextSub("ic", "BAG_UPDATE", -- Item text substitutions, triggered with the BAG_UPDATE event. Syntax: $(ic) for use on the action id of the button its on, $(ic item name) to show the count of a specific item
		function(class, item)
			--[[
			Behaviour:
				- If no item is specified the current actionslot will be used if possible
				- If an item is specified it will try to find out that item's count and display it on the button
			]]
			if(item ~= "") then
				local _, ItemCount, _, _, _, _, _, _, _ = FlexBar2:GetItemInfoByName(item); -- GetItemInfoByName, very usefull function I made that returns everything I need in relation with items. See the function declaration in the FlexBar2.lua file for details
				return ItemCount; -- Return
			else
				if(type(Model:GetButtonSetting(class.ButtonID, "Action").action) ~= "number") then return ""; end -- Make sure the current button has an ActionID linked to it (might support spell cooldowns later, but that ain't a priority) (note on this check: the button does not have to be an actionbutton to have an actionid, u can have macro buttons or whatever with actionid's as well, they will then react to cooldowns and such as if they were that button) (copy & paste from the above as i'm lazy)
				local ActCount = GetActionCount(Model:GetButtonSetting(class.ButtonID, "Action").action); -- Regular get action call using the action id on the button's ID
				if(IsConsumableAction(Model:GetButtonSetting(class.ButtonID, "Action").action) and ActCount > 1) then -- Only show anything on consumable actions AND if the count is higher then 1
					return ActCount;
				end
			end
			return ""; -- Fallback to return an empty string
		end);

	Controller:RegisterTextSub("binding", {Controller, "SetKeyBinding"}, -- Keybinding textsub
		function(class) -- KeyBinding text su, simply return the saved binging or an empty string as fallbackb
			return Model:GetButtonSetting(class.ButtonNumber, "BindingKey") or "";
		end);

	Controller:RegisterTextSub("id", {Controller, "SetButtonID"}, -- ButtonID text sub
		function(class) -- Button ID text sub, simply return the saved ID or an empty string as fallback
			return class.ButtonID
		end);

	Controller:RegisterTextSub("number", nil, -- Button number text sub
		function(class) -- Button number text sub, simply return the number
			return class.ButtonNumber or "";
		end);


	for _, v in ipairs(Controller.DefaultEventList) do -- Register the events that are saved in the DefaultEventList using ace2
		RegisterEvent(self, v, Controller.Events.OnEvent); -- See the RegisterEvent function declaration for more details on how to use this function
	end


	FlexBar2:ScheduleRepeatingEvent(function() event = "OnUpdate"; Controller.Events.OnEvent() end, 0.5); -- Shedule a fake OnUpdate event that triggers each 0.5 sec instead of each frame (to have less FPS loss)


	 -- FlexBar loads very late (at PLAYER_LOGIN, so be sure to have access to all API functions and user data) so it missed most, if not all events triggered during loading, so we trigger them manualy as we need them here
	-- Triggers ACTIONBAR_SLOT_CHANGED once for every button so the textures are set correctly and buffs light up green correctly
	-- Triggers UPDATE_SHAPESHIFT_FORM to have textures updated correctly if you are shapeshifted when your interface loads
 	for ButtonNumber = 1, Model:GetCustomSetting("MaxButtons") do
		arg1 = ButtonNumber;
		local class = Model:GetTempData("ButtonObject" .. ButtonNumber);
		Controller.Events:ACTIONBAR_SLOT_CHANGED();
		class:ApplyButtonTexture(Controller:GetIDTexture(class.ButtonID));
		this = {};
		this.GetClass = function() return class end
		Controller.Events.Button:UPDATE_SHAPESHIFT_FORM(true);
		this = nil;
		Controller.Events.Button:UpdateButtonText(class); -- this forces all text subs to set themselves
	end
	-- Triggers UNIT_INVENTORY_CHANGED to have equiped items light blue as intended
	arg1 = "player";
	Controller.Events:UNIT_INVENTORY_CHANGED()
	arg1 = nil;

	FlexBar2:TriggerEvent("FB2_Startup"); -- When this triggers FB2 has started but is not fully working yet, but all functions and events are available and ready to be hooked into.
end


function Controller:GenerateClass(ButtonNumber) -- Generates a class, should only be used upon loading or when increasing maxbuttons
	local class =  FlexBar2.View.Class:new(ButtonNumber); -- Create the class, the class will have *no* direct reference, but will be stored in the tempdata
	Model:SetTempData("ButtonObject" .. ButtonNumber, class); -- Save the class reference in the TempData table
	if(type(Model:GetAllButtonSettings(ButtonNumber)) ~= "table") then -- Check if the button already has saved settings, if it doesn't load default ones (mainly on the first startup)
		Model:SetAllButtonSettings(ButtonNumber, {
			X =  100 + (38 * (ButtonNumber - ((math.ceil(ButtonNumber / 12) - 1) * 12))), -- set X coordinate
			Y = -(100 + (38 * (math.ceil(ButtonNumber / 12)))), -- set Y coordinate
			Create = true, -- Defaults to false (but set to true for testing purpose)
			Show = true,-- Default to false (but set to true for testing purpose)
			Lock = false,-- Default to false
			IconLock = false, -- Default to false
			Active = true, -- Default to true
			BindingKey = "", -- Default key binding
			BindingMouseButton = nil, -- Default mouseclick the keybinding will trigger
			GroupType = "parent", -- Can be parent or child depending if the current button is the parent of a group or just the child of one
			Group = { ButtonNumber, }, -- Expected to always be a table containing always containing the button's parent first (or itself if it is a parent of course)
			Action = { type = "action", action = ButtonNumber, },
			Texture = nil, -- Overrides the default texture on a button if not nil
			Scale = 1, -- Default scale
			ButtonColor = {1, 1 ,1}, -- Default button color
			ShowCooldownTicker = "global", -- Global cooldown setting, if set to global, the ticker will only show for global cooldown, if set to true it will always show, if set to false it will never show
			StanceSetup = {}, -- Stancesetup, managed by the SetupStanceRemap function, see that function declaration for more details
			TextField1 = "", -- topleft text field
			TextField2 = "$(binding)", -- topright text field
			TextField3 = "$(id)", -- buttomright text field
			TextField4 = "$(ic)", -- buttomleft text field
			TextField5 = "$(cd)", -- mid text field
		});
	end

	Controller:SetButtonID(class, Model:GetButtonSetting(class.ButtonNumber, "ButtonID")); -- Set the button's button ID with its saved value (see that function declaration for details, once again)


	for i = 1, 5 do -- Loop through every textfield and use the SetButtonText function to set it (see that function declaration for details)
		Controller:SetButtonText(class, i, Model:GetButtonSetting(class.ButtonNumber, "TextField" .. i))
	end


	Controller:LoadButton(class); -- Load the button(see that function declaration for details)
	Controller:ApplyButtonAction(class); -- Set the button's attribute's using the saved values (see that function declaration for details, *again* :=))


	Model:GetTempData("ButtonObject" .. ButtonNumber):UpdateButton(); -- Update the button's texture
	Controller:SetKeyBinding(class, Model:GetButtonSetting(class.ButtonNumber, "BindingKey"), Model:GetButtonSetting(class.ButtonNumber, "BindingMouseButton")) -- Restore the saved keybinding
	for n,v in pairs(Model:GetButtonSetting(ButtonNumber, "StanceSetup")) do
		Controller:SetUpStanceRemap(class, n, v);
	end
end


function Controller:ApplyButtonAction(class) -- Set the button's attributes using saved values, this controls any of the button's behaviour that was limited by blizzard (you might want to hook this function at the FB2_Startup events to register your own attributes)
	if(class.Frame) then -- Only do anything if the Frame was already created, it always should, except if a third party addon fucks anything up, in case we print out a warning
		for n,v in pairs(Model:GetButtonSetting(class.ButtonID, "Action")) do -- Use the button's id instead of the button's action here, this results in everything getting remapped by itself instead of having to do crappy coding :)
			class.Frame.Button:SetAttribute(n, v); -- SetAttribute call on the frame, can only be called while out of combat (no combat check yet, planning to put one in later)
		end
	else
		FlexBar2:Print("Error: Button " , class.ButtonNumber, " Has no frame, skipping it"); -- Error msg if we got no frame (shldn't happen, but who knows ...)
	end
end

function Controller:RegisterTextSub(prefix, update, func)
	--[[
	this is used to register a text sub handler, the handler is put in a table and is automaticly called with the class and suffix as argument when on the text sub update event
	Arguments: 
		prefix: the 'name' of the textsubs (example: 'cd' if the user shld call $(cd) to access the sub)
		update: when to update the field containing that textsub should be updated
		func: the function to call when updating the textsub
	Saving: the handlers are NOT saved through session and have to be reregistered each time the UI loads
		]]
	local TextSubTable = Model:GetTempData("TextSubTable"); -- Get a local copy of the table containing all text subs
	local CustomEventTable = Model:GetTempData("CustomEventTable"); -- Get a local copy of the table containing events
	TextSubTable[prefix] = func; -- Put the func that updates the sub in the table
	local UpdateFunc = function(class) -- Special function that updates all the textfields of all buttons when it gets called
		for ButtonNumber = 1, Model:GetCustomSetting("MaxButtons") do -- Loop through all buttons, once again
			Controller.Events.Button:UpdateButtonText(Model:GetTempData("ButtonObject" .. ButtonNumber), prefix); -- Update the text labels on all buttons
		end
	end
	if(type(update) == "string") then -- If update is a string we just assume it contains the event name to react on
		Controller:AddCustomEvent(update, UpdateFunc);
	elseif(type(update) == "table") then -- If update is a table then we hook into the function that is pointed to in that table
		if(update[2]) then -- We hook a method
			FlexBar2:SecureHook(update[1], update[2], UpdateFunc);
		else -- We hook a global function
			FlexBar2:SecureHook(update[1], UpdateFunc);
		end
	end
	-- Save the new settings
	Model:SetTempData("TextSubTable", TextSubTable);
	Model:SetTempData("CustomEventTable", CustomEventTable);
end


function Controller:SetButtonText(class, field, text)
	--[[
	Sets the text of one of the given button's text fields
	Arguments:
		class: the object to set the text to (example: FlexBar2.Model:GetTempData("ButtonObject1");)
		field: the field id to set the text to (1 , 2, 3, 4 or 5)
		text: the text to set the field to (including subs)
	Saving: this function always saves the text through sessions
	Working: this parses the passed text string and saves the result, so instead of parsing each time the text sub gets updated we only parse now and loop through the result when we update the text sub
	]]
	Model:SetButtonSetting(class.ButtonNumber, "TextField" .. field, text); -- Save the button text in the database
	local PrefixList, SuffixList = {}, {}; -- Get 2 empty tables to put the prefix & suffix list in	
	for match in string.gmatch(Model:GetButtonSetting(class.ButtonNumber, "TextField" .. field), "$%((.-)%)") do -- Parse the text and loop through the results (detects valid textsubs)
		local _, _, prefix, suffix = string.find(match, "(%a*) ?(.*)"); -- reparse the result of the first parsing (seperates prefix and suffix in the sub)
		table.insert(PrefixList, prefix); -- Save the prefix in the list
		table.insert(SuffixList, suffix); -- Save the suffix in the list
	end
	Model:SetTempData("Button" .. class.ButtonNumber .. "TextSub" .. field .. "PrefixList", PrefixList); -- Save the new prefix list
	Model:SetTempData("Button" .. class.ButtonNumber .. "TextSub" .. field .. "SuffixList", SuffixList); -- Save the new suffix list
end

function Controller:SetButtonTexture(class, texture, savesetting)
	--[[
	sets the button texture to something new
	Arguments:
		class: the object to set the text to (example: FlexBar2.Model:GetTempData("ButtonObject1");)
		texture: the texture path of the new texture
	Saving: this functions accepts a third boolean argument, if set to true the texture will be saved through sessions by flexbar, if unset or set to false it will not
	Working: passes the new texture to the passed class's ApplyButtonTexture method and saves through sessions if told so
	]]
	class:UpdateButtonTexture(texture); -- Update the button's texture
	if(savesetting) then -- Check if we have to save our settings through sessions
		Model:SetButtonSetting(class.ButtonNumber, "Texture", texture); -- Save the setting
	end
end

function Controller:SetButtonScale(class, scale, savesetting)
	--[[
	Changes a button's scale
	Arguments:
		class: the object to set the text to (example: FlexBar2.Model:GetTempData("ButtonObject1");)
		scale: the new scale to set the button to
	Saving: this functions accepts a third boolean argument, if set to true the texture will be saved through sessions by flexbar, if unset or set to false it will not
	Working: passes the new scale to the passed class's ApplyButtonScale method and saves through sessions if told so
	]]
	class:ApplyButtonScale(scale); -- Update the button's scale
	if(savesetting) then -- Check if we have to save our settings through sessions
		Model:SetButtonSetting(class.ButtonNumber, "Scale", scale); -- Save the setting
	end
	
end

function Controller:SetButtonColor(class, r, g, b, savesetting) -- Always saves its settings through session
	--[[
	Changes the button's color
	Arguments:
		class: the object to set the text to (example: FlexBar2.Model:GetTempData("ButtonObject1");)
		r: the amount of red to use, as passed to SetVertexColor
		g: the amount of green to use, as passed to SetVertexColor
		b: the amount of blue to use, as passed to SetVertexColor
	Saving: this functions accepts a fifth boolean argument, if set to true the texture will be saved through sessions by flexbar, if unset or set to false it will not
	Working: passes the new color to the passed class's ApplyIconColor method and saves through sessions if told so
	]]
	class:ApplyIconColor(r, g, b); -- Update the button's icon color
	if(savesetting) then -- Check if we have to save our settings through sessions
		Model:SetButtonSetting(class.ButtonNumber, "ButtonColor", {r, g, b}); -- Save the setting
	end
end

function Controller:SetKeyBinding(class, BindingKey, BindingMouseButton, savesetting)
	--[[
	Registers a button's key binding
	Arguments:
		class: the object to set the text to (example: FlexBar2.Model:GetTempData("ButtonObject1");)
		BindingKey: The key to bind the button to as passed to SetBindingClick
		BindingMouseButton: the mouse button that will be simulated when using the keybinding
	Saving: this functions accepts a fourth boolean argument, if set to true the texture will be saved through sessions by flexbar, if unset or set to false it will not
	Working: passes the new color to the blizzard's SetBindingClick functions and saves through sessions if told so
	]]
	SetBindingClick(BindingKey, class.Frame.Button:GetName(), BindingMouseButton); -- Not performing sanity checks, expected to be valid syntax for the function
	if(savesetting) then
		Model:SetButtonSetting(class.ButtonNumber, "BindingKey", BindingKey); -- Save the parent's settings
		Model:SetButtonSetting(class.ButtonNumber, "BindingMouseButton", BindingMouseButton); -- Save the parent's settings
	end
end

function Controller:GetIDTexture(ButtonID)
	if(Model:GetButtonSetting(ButtonID, "texture")) then
		return Model:GetButtonSetting(ButtonID, "texture");
	elseif(Model:GetTempData("ButtonObject" .. ButtonID).Frame.Button:GetAttribute("type") == "action") then
		return GetActionTexture(Model:GetTempData("ButtonObject" .. ButtonID).Frame.Button:GetAttribute("action")) or nil;
	else
		return nil;
	end
end

function Controller:SetUpStanceRemap(class,StanceID, NewID, savesetting)
	--[[
	Registers a button's key binding
	Arguments:
		class: the object to set the text to (example: FlexBar2.Model:GetTempData("ButtonObject1");)
		StanceID: What Stance the switch has to be done on (list of possibilities bellow)
		NewID: the ID of the *button* to remap to
	Saving: this functions accepts a fourth boolean argument, if set to true the Remap will be saved and restore when logging back in, if set to false it'll only last till the next reloadUI
	Working: Generates a StateButton attribute to pass to the secure template. Then loops through all the attributes of the button it is suposed to remap to and passes them to the secure template too.
	Possible StanceIDs:
		Druid
			* 0 = UnShifted
			* 1 = Bear/Dire Bear Form
			* 2 = Aquatic Form
			* 3 = Cat Form
			* 4 = Travel Form
			* 5 = Moonkin or Tree Form
			* 6 = Flying form

		Rogue
			* 0 = Visible
			* 1 = Stealth

		Warrior
			* 1 = Battle Stance
			* 2 = Defensive Stance
			* 3 = Beserker Stance

		Priest
			* 0 = Normal
			* 1 = ShadowForm

		Other classes
			* Stance settings are ignored for all other classes
	]]
	local StateButton = ""; -- Need an empty string-type var to store data in
	for i = 0, GetNumShapeshiftForms() do -- Regenerate the StateButton attribute
		StateButton = StateButton .. i .. ":Stance"..i .. ";"; -- Add all possible stances in the state button attribute
	end
	FlexBar2:Debug("statebutton: " .. StateButton);
	class.Frame.Button:SetAttribute("statebutton", StateButton); -- Pass the statebutton attribute to the secure template, this attribute tells the secure template where it can find all the other attributes.
	for n, v in pairs(Model:GetButtonSetting(NewID, "Action")) do -- Loop through all the attributes of the button we're suposed to remap to so we can pass em all to the secure template
		FlexBar2:Debug(n .. "-" .. "Stance" .. StanceID, v);
		class.Frame.Button:SetAttribute(n .. "-" .. "Stance" .. StanceID, v); -- Pass the attribute to the secure template
	end
	class.Frame.Button:SetAttribute("ID-Stance" .. StanceID, NewID); -- Custom attribute used to find out whats the new ID on secure remaps
	if(savesetting) then
		FlexBar2:Debug("saving");
		local StanceSetup = FlexBar2.Model:GetButtonSetting(class.ButtonNumber, "StanceSetup"); -- Retrieve the old saved setup
		StanceSetup[StanceID] = NewID; -- Set the new saved attributes
		FlexBar2.Model:SetButtonSetting(class.ButtonNumber, "StanceSetup", StanceSetup); -- Save the new attributes
	end
end

function Controller:SetButtonID(class, ID) -- Always saved, regardless of settings
	if(type(ID) ~= "number") then ID = class.ButtonNumber; end
	class.ButtonID = ID;
	if(Model:GetButtonSetting(class.ButtonNumber, "ButtonID")) then -- Remove old settings if they exist
		local OldID = Model:GetButtonSetting(class.ButtonNumber, "ButtonID");
		local OldTable = Model:GetTempData("ButtonIDButtons" .. OldID);
		if(FlexBar2:RemoveByVal(OldTable, class.ButtonNumber)) then Model:SetTempData("ButtonIDButtons" .. OldID, OldTable); end
	end
	local NewTable = Model:GetTempData("ButtonIDButtons" .. ID);
	if(type(NewTable) ~= "table") then NewTable = {} end; -- Check if the table already exists, if it doesn't, make it, shld only ever be done on login, if its not during login, something is wrong
	table.insert(NewTable, class.ButtonNumber);
	Model:SetTempData("ButtonIDButtons" .. ID, NewTable);
	Model:SetButtonSetting(class.ButtonNumber, "ButtonID", ID);
	if(class.Frame) then -- Need the if check in case we're still starting up, in which case this would still fail
		class:UpdateButton();
		Controller:ApplyButtonAction(class);
	end 
end

function Controller:LoadButton(class) -- Called to load a button and apply all settings, the settings get applied regardless of current status, so use with care
	if(Model:GetButtonSetting(class.ButtonNumber, "Create")) then -- Create the button if our saved settings say so, if the button doesn't get created, we ignore all other settings
		Controller:Create(class);
		if(Model:GetButtonSetting(class.ButtonNumber, "Show")) then
			Controller:Show(class);
		else
			Controller:Hide(class);
		end; -- Hide/Show the button if our saved settings say so
		if(Model:GetButtonSetting(class.ButtonNumber, "Lock")) then -- Locking check, things need to be done both if button has to be locked or unlocked (as the default state is neither)
			Controller:Lock(class);
		else
			Controller:Unlock(class);
		end
		if(not Model:GetButtonSetting(class.ButtonNumber, "IconLock")) then
			Controller:IconUnlock(class);
		end
		-- Register the classic events we always need
		class.Frame.Button:SetScript("PreClick", Controller.Events.Button.PreClick);
		class.Frame.Button:SetScript("PostClick", Controller.Events.Button.PostClick);
		class.Frame.Button:SetScript("OnEvent", Controller.Events.OnEvent);
	end
end
function Controller:Create(class, savesetting)
	class:Create();
	class.Frame.GetClass = function() return class; end; -- This function is used when having the frame sent through instead of the class (Mainly when reacting on events) to find out what the class is
	class.Frame.Button.GetClass = function() return class; end; -- Same as the above but for the button
	class.Frame.Button.StateChanged = function(NewState) FlexBar2:Debug("StateChangedTo" .. NewState); self:GetClass():UpdateButton(); end;
	Model:GetTempData("HeaderFrame"):SetAttribute("addchild", class.Frame.Button); -- Register the button with the SecureStateHeader
	class.Frame.Button:SetAttribute("useparent-statebutton", true); -- tell the frame to listen to its parent
	if(savesetting) then
		Model:SetButtonSetting(class.ButtonNumber, "Create", true);
	end
end

function Controller:Destroy(class) -- Note on destroy: button objects and their frames will continue to exist untill the UI is reloaded, this just sets the button to not be created upon reloading and hides it
	class:Hide();
	Model:SetButtonSetting(class.ButtonNumber, "Create", false);
end

function Controller:Group(ParentClass, ChildClass) -- Always saved through sessions
	--[[
	Behaviour:
		- If the ParentClass is a child: then the ChildClass gets grouped with the Parent of the ParentClass
		- If the ChildClass is a child: the child gets removed from its current group and added to the group of the ParentClass
		- If both ParentClass & ChildClass are parents: Normal behaviour, ChildClass and all of its childs gets grouped to ParentClass directly
	--]]
	FlexBar2:Debug("Group: " .. ParentClass.ButtonNumber .. "," .. ChildClass.ButtonNumber);
	if(Model:GetButtonSetting(ParentClass.ButtonNumber, "GroupType") == "child") then -- The parent is currently the child of another group, so we're gonna group the ChildClass to the parent of the ParentClass (so we're calling this function again with new parameters)
		local ParentTable = Model:GetButtonSetting(ParentClass.ButtonNumber, "Group"); -- Get the parent's table
		return Controller:Group(ParentTable[1], ChildClass); -- Call this function and then return again as I cba to do conditionals
	end
	if(Model:GetButtonSetting(ChildClass.ButtonNumber, "GroupType") == "child") then
		Controller:UnGroup(ChildClass); -- Ungroup the Child away from its current group
		return Controller:Group(ParentClass, ChildClass); -- Call this function again now the child ain't grouped anymore, as again, i'm lazy !
	end
	local ParentTable = Model:GetButtonSetting(ParentClass.ButtonNumber, "Group");
	local ChildTable = Model:GetButtonSetting(ChildClass.ButtonNumber, "Group");
	-- Now time for the actual grouping code :), we're sure both ParentClass & ChildClass are actual parents now, so just inserting all the childs of ChildClass into the ParentClass child list will do
	for _,v in ipairs(ChildTable) do
		table.insert(ParentTable, v);
		Controller:Lock(Model:GetTempData("ButtonObject" .. v), true); -- Lock the childs so they can't move on their own
		Model:SetButtonSetting(v, "GroupType", "child");
		Model:SetButtonSetting(ChildClass.ButtonNumber, "Group", { ParentClass.ButtonNumber, }); -- Set the child's table to contain the parent
	end
	Model:SetButtonSetting(ParentClass.ButtonNumber, "Group", ParentTable); -- Save the parent's settings
end

function Controller:UnGroup(class) -- Always saved through sessions
	--[[
	Behaviour:
		- If a parent is passed: disband the group completely
		- If a child is passed: Remove that child from the group it is in
	--]]
	FlexBar2:Debug("UnGroup: " .. class.ButtonNumber .. "(" .. Model:GetButtonSetting(class.ButtonNumber, "GroupType") .. ")");
	if(Model:GetButtonSetting(class.ButtonNumber, "GroupType") == "parent") then
		local Table = FlexBar2:TableCopy(Model:GetButtonSetting(class.ButtonNumber, "Group"));
		for n, v in ipairs(Table) do
			if(n ~= 1) then Controller:UnGroup(Model:GetTempData("ButtonObject" .. v)); end -- Ungroup the child
		end
	elseif(Model:GetButtonSetting(class.ButtonNumber, "GroupType") == "child") then
		Model:SetButtonSetting(class.ButtonNumber, "GroupType", "parent"); -- transform the frame into a parent
		if(not Model:GetButtonSetting(Model:GetButtonSetting(class.ButtonNumber, "Group")[1], "Lock")) then -- Unlock the child if the parent was previously unlocked
			Controller:Unlock(class, true);
		end
		local ParentTable = Model:GetButtonSetting(Model:GetButtonSetting(class.ButtonNumber, "Group")[1], "Group");
		if(FlexBar2:RemoveByVal(ParentTable, class.ButtonNumber)) then Model:SetButtonSetting(Model:GetButtonSetting(class.ButtonNumber, "Group")[1], "Group", ParentTable); end -- Remove the child that we just ungrouped from the parents table
		Model:SetButtonSetting(class.ButtonNumber, "Group", { class.ButtonNumber, });
	end
end

function Controller:SaveCustomEvent(event, script) -- Always saves its settings through sessions
	if(type(event) == "string" and type(script) == "string") then
		local SavedCustomEvents = Model:GetCustomSetting("SavedCustomEvents");
		if(type(SavedCustomEvents) ~= "table") then SavedCustomEvents = {}; end
		if(type(SavedCustomEvents[event]) ~= "table") then SavedCustomEvents[event] = {} end
		table.insert(SavedCustomEvents[event], script);
		Model:SetCustomSetting("SavedCustomEvents", SavedCustomEvents);
	end
end

function Controller:DeleteSavedCustomEvent(event, id) -- Always saves its settings through sessions
	if(type(event) == "string" and type(id) == "number") then
		local SavedCustomEvents = Model:GetCustomSetting("SavedCustomEvents");
		if(type(SavedCustomEvents[event]) ~= "table") then SavedCustomEvents[event] = {} end
		table.remove(SavedCustomEvents[event], id);
		Model:SetCustomSetting("SavedCustomEvents", SavedCustomEvents);
	end
end

function Controller:AddCustomEvent(event, func) -- Custom events are NOT saved through sessions, so if you want to use custom events, the best thing to do is to add them to a script that triggers with ADDON_LOADED or use the flexbar2 event editor
	local CustomEventTable = Model:GetTempData("CustomEventTable");
	if(not CustomEventTable[event]) then
		CustomEventTable[event] = {};
	end
	if(not FlexBar2:ValIsInTable(Controller.DefaultEventList, event)) then
		if(not FlexBar2:IsEventRegistered(event) and event ~= "OnUpdate") then
			FlexBar2:RegisterEvent(event, Controller.Events.OnEvent);
		end
	end
	if(type(func) == "function") then
		table.insert(CustomEventTable[event], func);
	end
	Model:SetTempData("CustomEventTable", CustomEventTable);
end

function Controller:RemoveCustomEvent(event)
	if(FlexBar2:ValIsInTable(CustomEventTable, event) and not FlexBar2:ValIsInTable(Controller.DefaultEventList, event)) then
		local CustomEventTable = Model:GetCustomSetting("CustomEventTable");
		FlexBar2:RemoveByVal(CustomEventTable, event);
		if(FlexBar2:IsEventRegistered(event)) then
			FlexBar2:UnregisterEvent(event);
		end
		Model:SetCustomSetting("CustomEventTable", CustomEventTable);
	end
end

function Controller:Show(class, showstates, savesetting)
	-- the showing and hiding itself is controled by the SecureStateHeader, the class's :Show & :Hide method shld only take care of hiding & showing the other stuff that has to
	class:Show();
	class.Frame.Button:SetAttribute("hidestates", ATTRIBUTE_NOOP);
	class.Frame.Button:SetAttribute("showstates", "*");

	for _, v in ipairs(Controller.DefaultButtonEventList) do
		class.Frame.Button:RegisterEvent(v);
	end
	if(savesetting) then
		Model:SetButtonSetting(class.ButtonNumber, "Show", true);
	end
end

function Controller:Hide(class, savesetting)
	class:Hide();
	class.Frame.Button:SetAttribute("showstates", ATTRIBUTE_NOOP);
	class.Frame.Button:SetAttribute("hidestates", "*");
	for _, v in ipairs(Controller.DefaultButtonEventList) do
		class.Frame.Button:UnregisterEvent(v);
	end
	if(savesetting) then
			Model:SetButtonSetting(class.ButtonNumber, "Show", false);
	end
end


function Controller:Lock(class, savesetting)
	class:Lock();
	class.Frame:SetScript("OnMouseUp", nil);
	class.Frame:SetScript("OnMouseDown", nil);
	if(savesetting) then
			Model:SetButtonSetting(class.ButtonNumber, "Lock", true);
	end
end

function Controller:Unlock(class, savesetting)
	class:Unlock();
	class.Frame:SetScript("OnMouseUp", Controller.Events.Frame.StopDragging);
	class.Frame:SetScript("OnMouseDown", Controller.Events.Frame.StartDragging);
	if(savesetting) then
			Model:SetButtonSetting(class.ButtonNumber, "Lock", false);
	end
end

function Controller:IconLock(class, savesetting)
	class.Frame.Button:RegisterForDrag();
	class.Frame.Button:SetScript("OnDragStart", nil);
	class.Frame.Button:SetScript("OnReceiveDrag", nil);
	if(savesetting) then
			Model:SetButtonSetting(class.ButtonNumber, "IconLock", true);
	end
end

function Controller:IconUnlock(class, savesetting)
	class.Frame.Button:RegisterForDrag("LeftButton");
	class.Frame.Button:SetScript("OnDragStart", Controller.Events.Button.DragAction);
	class.Frame.Button:SetScript("OnReceiveDrag", Controller.Events.Button.ReceiveAction);
	if(savesetting) then
			Model:SetButtonSetting(class.ButtonNumber, "IconLock", false);
	end
end

Controller.Bindings = {};
Controller.Events = {};
Controller.Events.Frame = {};
function Controller.Events.Frame:StartDragging()
	local _, _, _, pX, pY = this:GetPoint("TOPLEFT");
	for n, v in ipairs(Model:GetButtonSetting(this:GetClass().ButtonNumber, "Group")) do -- Find out if the frame has childs, if it does, we take em all with us
		if(n ~= 1) then -- Don't do anything if v is 1, as thats the parent itself
			local _, _, _, cX, cY = Model:GetTempData("ButtonObject" .. v).Frame:GetPoint("TOPLEFT");
			-- Unfortunatly, we can't drag multiple frames at once, so we anchor everything to the parent, and drag the parent, we unanchor everything when we're done dragging
			Model:GetTempData("ButtonObject" .. v).Frame:ClearAllPoints();
			Model:GetTempData("ButtonObject" .. v).Frame:SetPoint("TOPLEFT", this, "TOPLEFT", (cX - pX), (cY - pY));
			FlexBar2:Debug((cX - pX), (cY - pY));
		end
	end
	this:GetClass():StartDragging();
end

function Controller.Events.Frame:StopDragging()
	this:GetClass():StopDragging();
	local _, _, _, pX, pY = this:GetPoint();
	for _, v in ipairs(Model:GetButtonSetting(this:GetClass().ButtonNumber, "Group")) do -- Find out if the frame has childs, if it does, we stop dragging em all
		if(Model:GetButtonSetting(v, "GroupType") == "parent") then -- Different behaviour wether the frame is a parent or not
			Model:SetButtonSetting(Model:GetTempData("ButtonObject" .. v).ButtonNumber, "X", pX);
			Model:SetButtonSetting(Model:GetTempData("ButtonObject" .. v).ButtonNumber, "Y", pY);
		elseif(Model:GetButtonSetting(v, "GroupType") == "child") then
			local _, _, _, cX, cY = Model:GetTempData("ButtonObject" .. v).Frame:GetPoint();
			FlexBar2:Debug("(" .. pX .. ":" .. pY .. "),(" .. cX .. ":" .. cY .. ")");
			Model:GetTempData("ButtonObject" .. v).Frame:ClearAllPoints();
			Model:GetTempData("ButtonObject" .. v).Frame:SetPoint("TOPLEFT", "UIParent", "TOPLEFT", (pX + cX), (pY + cY));
			Model:SetButtonSetting(Model:GetTempData("ButtonObject" .. v).ButtonNumber, "X", (pX + cX));
			Model:SetButtonSetting(Model:GetTempData("ButtonObject" .. v).ButtonNumber, "Y", (pY + cY));
		end
	end
end

Controller.Events.Button = {}

function Controller.Events:OnEvent(buttonoverride, carg1, carg2, carg3, carg4, carg5, carg6) -- event handler, will redirect any event it gets to the correct method and update button text if needed
	--[[
	Behaviour:
		- Custom args (mainly coming from Ace) are passed to the called functions
		- Global args (set by blizz event) are left alone
		- First the handler in the Controller.Events table gets called, if available
		- Then the handler in Controller.Events.Button table gets called, if available
		- Then the handler in the CustomEventTable gets called, if available
	]]
		if((this and this.GetClass) or buttonoverride) then -- if this and this.getclass exist we deal with a button event, so we call the button functions but NOT the regular or custom ones (we also call events that explicitly override that setting
			if(type(Controller.Events.Button[event]) == "function") then Controller.Events.Button[event](self, carg1, carg2, carg3, carg4, carg5, carg6); end -- Button event caller
			if(this and this.GetClass and this:GetClass().ButtonNumber == 1 and event ~= "OnUpdate") then
				FlexBar2:Debug("OnEvent (Button): "..event);
			end
		else -- This is a regular event, call only the regular and custom functions and NOT the button dependenant ones
			if(event ~= "OnUpdate") then FlexBar2:Debug("OnEvent (" ..event .. ")" .. ":", carg1, ":", carg2, ":", carg3, ":", carg4, ":", carg5, ":", carg6); end
			if(type(Controller.Events[event]) == "function") then Controller.Events[event](self, carg1, carg2, carg3, carg4, carg5, carg6); end -- Regular event code
			if(type(Model:GetTempData("CustomEventTable")[event]) == "table") then -- Custom event table loop
				for n,v in ipairs(Model:GetTempData("CustomEventTable")[event]) do
					if(type(v) == "function") then v(self, carg1, carg2, carg3, carg4, carg5, carg6); end
				end
			end
		end
end

function Controller.Events:OnUpdate()
	-- Localizing functios called way too often for performance (the gain is marginal though, but any gain is worth it in OnUpdate)
	local GetButtonSetting = Model.GetButtonSetting;
	local GetTempData = Model.GetTempData;
	local IsUsableAction = IsUsableAction;
	local IsActionInRange = IsActionInRange;
	local GetActionCooldown = GetActionCooldown;

	for i = 1, Model:GetCustomSetting("MaxButtons") do
		local ActionID = GetButtonSetting(self, i, "Action").action;
		local _, ActDur, _ = GetActionCooldown(ActionID);
		local ActionCount = GetActionCount(ActionID);
		local SpellUsable = (IsUsableAction(ActionID) == 1 and IsActionInRange(ActionID) ~= 0 and ActDur <= 1.5);
		if(ActionID == nil) then break; end
		local SpellWasUsable = GetTempData(self, "Action" .. ActionID .. "IsUsable");
		if(SpellUsable ~= SpellWasUsable) then -- Trigger the event
			FlexBar2:TriggerEvent("FB2_ActionIsUsableStateChanged", FlexBar2, true, ActionID, i, SpellUsable);
			Model:SetTempData("Action" .. ActionID .. "IsUsable", SpellUsable);
		end
	end
end

function Controller.Events:ACTIONBAR_SLOT_CHANGED() -- This should normaly be useable for channeled spells, but it doesn't seem to always trigger, so only using it for buffs
	-- Localize funcs used in loops to increase performance
	local GetTempData = Model.GetTempData;
	local GetButtonSetting = Model.GetButtonSetting;

	if(arg1 == 0) then return; end -- gets triggered with arg1 being 0 when zoning, return if that happens
--[[	for i = 1, Model:GetCustomSetting("MaxButtons") do -- Update all the buttons with the 
		local class = GetTempData(self, "ButtonObject" .. i);
		class:UpdateButton();
		class:ApplyButtonTexture(Controller:GetIDTexture(class.ButtonID));
	end]]
	if(GetTempData(self, arg1 .. "Equipped") == false) then
		for i = 1, Model:GetCustomSetting("MaxButtons") do
			if(GetButtonSetting(self, i, "Action").action == arg1) then
				GetTempData(self, "ButtonObject" .. i):UnCheck();
			end
		end
	end
	for i = 1, 16 do -- Check for buffs to light green
		local _, _, buff = UnitBuff("player", i);
		if(buff ~= nil and buff == GetActionTexture(arg1)) then
			if(GetActionTexture(arg1)) then FlexBar2:Debug("BUFF: ".. GetActionTexture(arg1)); end
			for _, v in ipairs(GetTempData(self, "ButtonIDButtons" .. arg1)) do -- Update all the buttons that have the buttonid that was updated
				FlexBar2:Debug("UPDATE: " .. v);
				GetTempData(self, "ButtonObject" .. v):CheckBuffed();
			end
			Model:SetTempData(arg1 .. "Buffed", true);
			return;
		end
	end
	Model:SetTempData(arg1 .. "Buffed", false);
end

function Controller.Events:UNIT_INVENTORY_CHANGED()
	if(arg1 == "player") then
		-- Localize funcs used in loops to increase performance
		local GetButtonSetting = Model.GetButtonSetting;
		local SetTempData = Model.SetTempData;
		local GetTempData = Model.GetTempData;
		local IsEquippedAction = IsEquippedAction;

		for i = 1, Model:GetCustomSetting("MaxButtons") do
			local ActionID = GetButtonSetting(self, GetTempData(self, "ButtonObject" .. i).ButtonID, "Action").action
			if(IsEquippedAction(ActionID)) then
				SetTempData(self, ActionID .. "Equipped", true);
				for _, v in ipairs(GetTempData(self, "ButtonIDButtons" .. i)) do -- Update all the buttons that have the buttonid that was updated
					GetTempData(self, "ButtonObject" .. v):CheckEquipped();
				end
			else
				SetTempData(self, ActionID .. "Equipped", false);
				if(GetTempData(self, ActionID .. "Buffed") == false) then
					for _, v in ipairs(GetTempData(self, "ButtonIDButtons" .. i)) do -- Update all the buttons that have the buttonid that was updated
						GetTempData(self, "ButtonObject" .. v):UnCheck();
					end
				end
			end
		end
	end
end


function Controller.Events.Button:StopDragging() -- Called when button dragging stop is detected
	FlexBar2:Debug("StopDragging");
	for i = 1, Model:GetCustomSetting("MaxButtons") do
		Controller:ApplyButtonAction(Model:GetTempData("ButtonObject" .. i));
	end
end


function Controller.Events.Button:PreClick()
	MacroFrame_SaveMacro();
	r, g, b = this:GetClass().Frame.Button:GetCheckedTexture():GetVertexColor(); -- Save the old settings of the button to restore em after we're done
	if((r == nil or g == nil or b == nil) or (r > 0.99 or g > 0.99 or b > 0.99)) then this:GetClass():UnCheck(); end
end

function Controller.Events.Button:PostClick()
	if(IsCurrentAction(this:GetClass().ButtonID)) then
		this:GetClass():Check();
		this:RegisterEvent("UNIT_SPELLCAST_START"); -- this event has to be registerred for spells like flamestrike that don't get cast immediately
		this:RegisterEvent("UNIT_SPELLCAST_FAILED"); -- Register the spellcast_failed event in case the casting fails
		this:RegisterEvent("UNIT_SPELLCAST_STOP"); -- Register the spellcast_stop event for this button for when the spellcast stops to uncheck it
		this:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START"); -- ACTIONBAR_SLOT_CHANGED doesn't seem to trigger for all channeled spells, so I won't rely on it, i'll use this instead :)
		this:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP");
		this:RegisterEvent("CURRENT_SPELL_CAST_CHANGED");
	end
end

function Controller.Events.Button:DragAction()
	FlexBar2:Debug("DragAction");
	PickupAction(Model:GetButtonSetting(this:GetClass().ButtonID, "Action").action); -- Put the action on the mouse and remove it from the button
	for ButtonNumber = 1, Model:GetCustomSetting("MaxButtons") do
		local class = Model:GetTempData("ButtonObject" ..ButtonNumber);
		class:ApplyButtonTexture(Controller:GetIDTexture(class.ButtonID));
		Controller.Events.Button:UpdateButtonText(class);
		if(not HasAction(Model:GetButtonSetting(class.ButtonID, "Action").action)) then
			class:UnCheck();
		end
	end
end

function Controller.Events.Button:ReceiveAction()
	FlexBar2:Debug("ReceiveAction");
	PlaceAction(Model:GetButtonSetting(this:GetClass().ButtonID, "Action").action); -- Place the action on the button
	for ButtonNumber = 1, Model:GetCustomSetting("MaxButtons") do
		local class = Model:GetTempData("ButtonObject" ..ButtonNumber);
		class:ApplyButtonTexture(Controller:GetIDTexture(class.ButtonID));
		Controller.Events.Button:UpdateButtonText(class);
		if(not HasAction(Model:GetButtonSetting(class.ButtonID, "Action").action)) then
			class:UnCheck();
		end
	end
end

function Controller.Events.Button:UpdateButtonText(class, forceprefix)
	for i = 1, 5 do
		local str = Model:GetButtonSetting(class.ButtonNumber, "TextField" .. i);
		local PrefixList = Model:GetTempData("Button" .. class.ButtonNumber .. "TextSub" .. i .. "PrefixList");
		local SuffixList = Model:GetTempData("Button" .. class.ButtonNumber .. "TextSub" .. i .. "SuffixList");
		local Update;
		for n, v in ipairs(PrefixList) do
			if(v == forceprefix or not forceprefix) then
				local suffix = SuffixList[n] or "";
				if(Model:GetTempData("TextSubTable")[v]) then
					str = string.gsub(str, "$%(" .. v .. "( ?)".. suffix .. "%)", Model:GetTempData("TextSubTable")[v](class, suffix));
					Update = true;
				end
			end
		end
		if(Update) then
			class:SetButtonText(i, str);
		end
	end
end

function Controller.Events.Button:FB2_ActionIsUsableStateChanged(ActionID, ButtonNumber, SpellUsable)
	--[[
	This event is triggered when an action button that was not usable becomes usable or vice versa (this is triggered quite often in combat, as global cooldown triggers it too) (this is also triggered once per button upen login)
	The following args are passed for this event:
		1: the Action ID of the button that changed
		2: the flexbar button number that the spell is on (if the spell is on multiple buttons it will be fired multiple times)
		3: true if the spell is usable now, false if its not
	]]
	if(SpellUsable) then
		Model:GetTempData("ButtonObject" .. ButtonNumber):EnableAction();
	else
		Model:GetTempData("ButtonObject" .. ButtonNumber):DisableAction();
	end
end

function Controller.Events.Button:UNIT_SPELLCAST_FAILED() -- This does exactly the same as SPELLCAST_STOP, so we'll just call it
	if(arg1 == "player") then Controller.Events.Button:UNIT_SPELLCAST_STOP(); end
end

function Controller.Events.Button:UNIT_SPELLCAST_START()
	if(arg1 == "player") then this:GetClass():Check(); end
end

function Controller.Events.Button:UNIT_SPELLCAST_STOP()
	if(arg1 == "player") then
		this:GetClass():UnCheck();
		this:UnregisterEvent("UNIT_SPELLCAST_FAILED");
		this:UnregisterEvent("UNIT_SPELLCAST_STOP");
		this:UnregisterEvent("UNIT_SPELLCAST_START");
		this:UnregisterEvent("CURRENT_SPELL_CAST_CHANGED");
		this:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_START");
		this:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_STOP");
		if((r ~= nil and g ~= nil or b ~= nil) and (r < 0.99 or g < 0.99 or b < 0.99)) then -- this ain't always exactly one, so i'll use < 0.99 instead
			FlexBar2:Debug("Restoring");
			this:GetClass():UnCheck();
			this:GetClass():Check(r, g, b);
		end
		r = nil;
		g = nil;
		b = nil;
	end
end

function Controller.Events.Button:UNIT_SPELLCAST_CHANNEL_START() -- This event doesn't pass thru ace2
	if(arg1 == "player") then
		this:GetClass():CheckChanneling();
		this:UnregisterEvent("UNIT_SPELLCAST_FAILED");
		this:UnregisterEvent("UNIT_SPELLCAST_STOP");
		this:UnregisterEvent("UNIT_SPELLCAST_CHANNEL_START");
		this:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP");
	end
end

function Controller.Events.Button:UNIT_SPELLCAST_CHANNEL_STOP() -- This event doesn't pass thru ace2
	if(arg1 == "player") then Controller.Events.Button:UNIT_SPELLCAST_STOP(); end
end

function Controller.Events.Button:CURRENT_SPELL_CAST_CHANGED() -- This event doesn't pass thru ace2
	if(arg1 == "player") then
		this:GetClass():UnCheck();
		this:UnregisterEvent("CURRENT_SPELL_CAST_CHANGED");
	end
end

function Controller.Events.Button:ACTIONBAR_UPDATE_COOLDOWN()
	if(type(Model:GetButtonSetting(this:GetClass().ButtonID, "Action").action) ~= "number") then return; end
	local start, duration, enable = GetActionCooldown(Model:GetButtonSetting(this:GetClass().ButtonID, "Action").action);
	if((this:GetClass().LastCooldown.start ~= start or this:GetClass().LastCooldown.duration ~= duration)) then
		if(Model:GetButtonSetting(this:GetClass().ButtonNumber, "ShowCooldownTicker") == "global") then
			if(duration <= 1.5) then
				this:GetClass():UpdateCooldown(start, duration, enable);
			end
		elseif(Model:GetButtonSetting(this:GetClass().ButtonNumber, "ShowCooldownTicker") == true) then
			this:GetClass():UpdateCooldown(start, duration, enable);
		end
	end
end

function Controller.Events.Button:UPDATE_SHAPESHIFT_FORM()
	local class = this:GetClass();
	if(class.Frame.Button:GetAttribute("statebutton")) then -- Only do anything if the button is using the secure state header at all
		-- We'll assume that states were setup my way (aka, using the SetUpStanceRemap function or one with the same syntax)
		local NewActionID = class.Frame.Button:GetAttribute("ID-Stance" .. GetShapeshiftForm(true));
		FlexBar2:Debug("Remap detected: New ID is '" .. NewActionID .. "'");
		if(NewActionID) then -- We can only update the texture if we were able to find out the new action ID (so this means that we will NOT update any modified textures, and that this may override nonsaved textures)
			class.ButtonID = NewActionID;
			class:ApplyButtonTexture(Controller:GetIDTexture(NewActionID)); -- This lovely method does all the hard work for me
		end
	end
end
