--[[
Name: FlexBar 2
Revision: $Rev: SVN $
Author(s): Cncfanatics (cvcncfanatic@runemasters.net)
    Dhaggar (isheridan@savagevines.com)
Website: http://flexdev.mmorpgaming.com
Documentation: http://flexdev.mmorpgaming.com/index.php?n=Main.Docs
Description: Makes your 120 actionbuttons incredibly versatile
Dependencies: AceLibrary
]]

-- WARNING: Calling any method of self.Frame class without sanity checks may cause a LOT of overhead and weird behaviour, like having the button to be visualy locked but still reacting to dragging
-- everything in self.Frame file only changes the LOOKS of a button not the events or anything else

FlexBar2.View = {};

FlexBar2.View.Cache = {};

FlexBar2.View.Class = FlexBar2.OO.Class();
function FlexBar2.View.Class.prototype:init(ButtonNumber)
	FlexBar2.View.Class.super.prototype.init(self); -- Init the class
	self.ButtonNumber = ButtonNumber;
end

function FlexBar2.View.Class.prototype:Create()
	if(table.getn(FlexBar2.View.Cache) > 0) then -- Check if the buttoncache contains something, if it does we'll use that button instead of generating a new one, also, for the sake of myself, i'll assume that every button put in here is a valid one
		self.Frame = table.remove(FlexBar2.View.Cache); -- Grab a frame from buttoncache
		self:SetButtonTexture(); -- Logicly, I would do self.Frame outside self.Frame if clause, but sadly, I can't, as button creation itself requires self.Frame to be done too
	else
		-- Button parent frame creation (note: we don't need to use :SetUserPlaced(false) as button caching does NOT work with CreateFrame'd frames xD
		self.Frame = CreateFrame("Frame", nil, UIParent);
		self.Frame:SetFrameStrata("LOW");
		self.Frame:SetWidth(10);
		self.Frame:SetHeight(10);
		self.Frame:ClearAllPoints();
		self.Frame:SetPoint("TOPLEFT", "UIParent", "TOPLEFT", FlexBar2.Model:GetButtonSetting(self.ButtonNumber, "X"), FlexBar2.Model:GetButtonSetting(self.ButtonNumber, "Y"));

		-- Frame settings
		self.Frame:EnableMouse(true);
		self.Frame:SetMovable(true);
		self.Frame:SetToplevel(true);
		self.Frame:SetClampedToScreen(false);

		--dragging handle creation
		self.Frame.Handle = self.Frame:CreateTexture(nil, "ARTWORK");
		self.Frame.Handle:SetTexture("Interface\\Tooltips\\UI-StatusBar-Border");
		self.Frame.Handle:SetAllPoints(self.Frame);

		--Actual button creation (note: not setting a NormalTexture here, an apart function does that which is called later on [at the icon creation part])
		self.Frame.Button = CreateFrame("CheckButton", "FB2_BTN_" .. self.ButtonNumber, self.Frame, "SecureActionButtonTemplate"); -- Need to set a frame name here for SetBindingClick to work, sigh ...
		self.Frame.Button:SetWidth(36);
		self.Frame.Button:SetHeight(36);
		self.Frame.Button:ClearAllPoints();
		self.Frame.Button:SetPoint("TOPLEFT", self.Frame, "TOPLEFT", 10, -10);
		self.Frame.Button:SetPushedTexture("Interface\\Buttons\\UI-Quickslot-Depress");
		self.Frame.Button:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square");
		self.Frame.Button:GetHighlightTexture():SetBlendMode("ADD");
		self.Frame.Button:SetCheckedTexture("Interface\\Buttons\\CheckButtonHilight");
		self.Frame.Button:GetCheckedTexture():SetBlendMode("ADD");

		-- Text1 fontstring creation
		self.Frame.Button.Text1 = self.Frame.Button:CreateFontString(nil, "ARTWORK", "NumberFontNormal");
		self.Frame.Button.Text1:SetJustifyH("LEFT");
		self.Frame.Button.Text1:SetPoint("TOPLEFT", self.Frame.Button, "TOPLEFT");
		self.Frame.Button.Text1:SetPoint("TOPRIGHT", self.Frame.Button, "TOPRIGHT");

		-- Text2 fontstring creation
		self.Frame.Button.Text2 = self.Frame.Button:CreateFontString(nil, "ARTWORK", "NumberFontNormal");
		self.Frame.Button.Text2:SetJustifyH("RIGHT");
		self.Frame.Button.Text2:SetPoint("TOPLEFT", self.Frame.Button, "TOPLEFT");
		self.Frame.Button.Text2:SetPoint("TOPRIGHT", self.Frame.Button, "TOPRIGHT");

		-- Text3 fontstring creation
		self.Frame.Button.Text3 = self.Frame.Button:CreateFontString(nil, "ARTWORK", "NumberFontNormal");
		self.Frame.Button.Text3:SetJustifyH("RIGHT");
		self.Frame.Button.Text3:SetPoint("BOTTOMLEFT", self.Frame.Button, "BOTTOMLEFT");
		self.Frame.Button.Text3:SetPoint("BOTTOMRIGHT", self.Frame.Button, "BOTTOMRIGHT");

		-- Text4 fontstring creation
		self.Frame.Button.Text4 = self.Frame.Button:CreateFontString(nil, "ARTWORK", "NumberFontNormal");
		self.Frame.Button.Text4:SetJustifyH("LEFT");
		self.Frame.Button.Text4:SetPoint("BOTTOMRIGHT", self.Frame.Button, "BOTTOMRIGHT");
		self.Frame.Button.Text4:SetPoint("BOTTOMLEFT", self.Frame.Button, "BOTTOMLEFT");
		
		-- Text5 fontstring creation
		self.Frame.Button.Text5 = self.Frame.Button:CreateFontString(nil, "ARTWORK", "NumberFontNormal");
		self.Frame.Button.Text5:SetJustifyH("CENTER");
		self.Frame.Button.Text5:SetPoint("CENTER");

		-- Icon creation
		self.Frame.Button.Icon = self.Frame.Button:CreateTexture(nil, "BACKGROUND");
		self.Frame.Button.Icon:SetAllPoints(self.Frame.Button);

		-- Cooldown ticker creation
		self.Frame.Button.Cooldown = CreateFrame("Model", nil, self.Frame.Button, "CooldownFrameTemplate");
		self.LastCooldown = {};

		-- NormalTexture setup, Can only do self.Frame now because SetButtonTexture requires .Icon to be declared (declared it above)
		self:UpdateButton();
		self.Frame.Button:GetNormalTexture():SetWidth(64);
		self.Frame.Button:GetNormalTexture():SetHeight(64);
		self.Frame.Button:GetNormalTexture():ClearAllPoints();
		self.Frame.Button:GetNormalTexture():SetPoint("CENTER", 1, -1);

		-- Button settings
		self.Frame.Button:EnableMouse(true);
		self.Frame.Button:SetClampedToScreen(true);

	end
end

function FlexBar2.View.Class.prototype:UpdateButton()
	self:UnCheck();
	self:ApplyButtonScale();
	self:ApplyButtonTexture();
end

function FlexBar2.View.Class.prototype:UpdateCooldown(start, duration, enable)
	CooldownFrame_SetTimer(self.Frame.Button.Cooldown, start, duration, enable);
end
function FlexBar2.View.Class.prototype:ApplyButtonScale(scale)
	self.Frame.Button:SetScale(scale or FlexBar2.Model:GetButtonSetting(self.ButtonNumber, "Scale"));
end

function FlexBar2.View.Class.prototype:ResetIconColor()
	local r, g, b = self.Frame.Button.Icon:GetVertexColor();
	self.Frame.Button.Icon:SetVertexColor(1 / r, 1 / g, 1 / b);
end

function FlexBar2.View.Class.prototype:ApplyIconColor(r, g, b)
	self.Frame.Button.Icon:SetVertexColor(
		r or FlexBar2.Model:GetButtonSetting(FlexBar2.Model:GetButtonSetting(self.ButtonNumber, "ButtonID"), "ButtonColor")[1],
		g or FlexBar2.Model:GetButtonSetting(FlexBar2.Model:GetButtonSetting(self.ButtonNumber, "ButtonID"), "ButtonColor")[2],
		b or FlexBar2.Model:GetButtonSetting(FlexBar2.Model:GetButtonSetting(self.ButtonNumber, "ButtonID"), "ButtonColor")[3]
	);
end

function FlexBar2.View.Class.prototype:EnableAction()
	self:ResetIconColor();
	self:ApplyIconColor();
end

function FlexBar2.View.Class.prototype:DisableAction()
	self.Frame.Button.Icon:SetVertexColor(0.5, 0.5, 0.5);
end

function FlexBar2.View.Class.prototype:SetButtonText(field, text)
	self.Frame.Button["Text" .. field]:SetText(text);
end

function FlexBar2.View.Class.prototype:ApplyButtonTexture(texture)
	self.Frame.Button.Icon:SetTexture(texture);
	if(texture) then
		self.Frame.Button:SetNormalTexture("Interface\\Buttons\\UI-Quickslot2");
	else
		self.Frame.Button:SetNormalTexture("Interface\\Buttons\\UI-Quickslot");
	end
end

function FlexBar2.View.Class.prototype:Show()
	self.Frame.Button:Show();
	if(not FlexBar2.Model:GetButtonSetting(self.ButtonNumber, "Lock")) then
		self:Unlock();
	end
end

function FlexBar2.View.Class.prototype:Hide()
	self.Frame.Button:Hide();
	self:Lock();
end

function FlexBar2.View.Class.prototype:Lock()
	self.Frame:Hide();
end

function FlexBar2.View.Class.prototype:Unlock()
	if(self.Frame.Button:IsShown()) then self.Frame:Show(); end
end

function FlexBar2.View.Class.prototype:StopDragging()
	if (self.Frame.isMoving) then
		self.Frame:StopMovingOrSizing();
		self.Frame.isMoving = false;
	end
end

function FlexBar2.View.Class.prototype:StartDragging()
	self.Frame:StartMoving();
	self.Frame.isMoving = true;
end

function FlexBar2.View.Class.prototype:CheckChanneling()
	self:Check(0.5, 0, 0.5);
end

function FlexBar2.View.Class.prototype:CheckBuffed()
	self:Check(0, 0.5, 0);
end

function FlexBar2.View.Class.prototype:CheckEquipped()
	self:Check(0, 0, 0.5);
end

function FlexBar2.View.Class.prototype:Check(r, g, b)
	self.Frame.Button:SetChecked(true);
	if(r and g and b) then
		self.Frame.Button:GetCheckedTexture():SetVertexColor(r, g, b);
	end
end

function FlexBar2.View.Class.prototype:UnCheck()
	self.Frame.Button:SetChecked(false);
	self.Frame.Button:GetCheckedTexture():SetVertexColor(1, 1, 1);
end

function FlexBar2.View.Class.prototype:Push()
	self.Frame.Button:SetButtonState("PUSHED");
end

function FlexBar2.View.Class.prototype:Release()

	self.Frame.Button:SetButtonState("NORMAL");
end
